/*
 * Copyright (C) 2021 Anton Romanov - All Rights Reserved
 * You may use, distribute and modify this code, please write to: romanov73@gmail.com.
 *
 */

package ru.ulstu.services;

import org.springframework.stereotype.Service;
import ru.ulstu.models.ModelingResult;
import ru.ulstu.models.TimeSeries;
import ru.ulstu.models.exceptions.ModelingException;
import ru.ulstu.score.ScoreMethod;
import ru.ulstu.score.Smape;
import ru.ulstu.tsMethods.MethodParameter;
import ru.ulstu.tsMethods.TimeSeriesMethod;
import ru.ulstu.tsMethods.TimeSeriesMethodParamValue;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;

@Service
public class MethodParamBruteForce {
    private final int DEFAULT_THREAD_COUNT = 50;
    private final List<TimeSeriesMethod> methods;
    private final ScoreMethod scoreMethod = new Smape();
    private final ExecutorService executors = Executors.newFixedThreadPool(DEFAULT_THREAD_COUNT);

    public MethodParamBruteForce(List<TimeSeriesMethod> methods) {
        this.methods = methods;
    }

    public TimeSeries getForecast(TimeSeries timeSeries) {
        throw new RuntimeException("Not implemented");
    }

    public TimeSeries getForecastWithOptimalLength(TimeSeries timeSeries) {
        throw new RuntimeException("Not implemented");
    }

    public ModelingResult getSmoothedTimeSeries(TimeSeries timeSeries) throws ModelingException {
        List<Future<ModelingResult>> results = new ArrayList<>();
        List<ModelingResult> results2 = new CopyOnWriteArrayList<>();
        try {
            for (TimeSeriesMethod method : methods) {
                List<List<TimeSeriesMethodParamValue>> availableParametersValues = getAvailableParametersValues(method.getAvailableParameters());
                for (List<TimeSeriesMethodParamValue> parametersValues : availableParametersValues) {
                    results.add(executors.submit(() -> {
                        TimeSeriesMethod methodInstance = method.getClass().getDeclaredConstructor().newInstance();
                        TimeSeries model = methodInstance.createFor(timeSeries)
                                .setAvailableParameters(parametersValues)
                                .getModel();
                        return new ModelingResult(model, parametersValues, scoreMethod.getScore(timeSeries, model), methodInstance);
                    }));
                }
            }
            results.forEach(modelingResultFuture -> {
                try {
                    modelingResultFuture.get();
                } catch (Exception e) {
                    e.printStackTrace();
                }
            });
            for (Future<ModelingResult> futureModelingResult : results) {
                results2.add(futureModelingResult.get());
            }
            return results2.stream()
                    .min(Comparator.comparing(modelingResult -> modelingResult.getScore().getDoubleValue()))
                    .orElse(null);
        } catch (Exception e) {
            throw new ModelingException(e.getMessage());
        }
    }

    private List<List<TimeSeriesMethodParamValue>> getAvailableParametersValues(List<MethodParameter> availableParameters) {
        List<List<TimeSeriesMethodParamValue>> result = new ArrayList<>();
        Map<MethodParameter, Integer> parameterOffset = new TreeMap<>();
        Map<MethodParameter, List<Number>> parameterValues = new TreeMap<>();
        for (MethodParameter methodParameter : availableParameters) {
            parameterOffset.put(methodParameter, 0);
            parameterValues.put(methodParameter, methodParameter.getAvailableValues());
        }
        while (!isAllValuesUsed(parameterOffset, parameterValues)) {
            List<TimeSeriesMethodParamValue> resultRow = new ArrayList<>();
            for (MethodParameter methodParameter : parameterOffset.keySet()) {
                resultRow.add(new TimeSeriesMethodParamValue(methodParameter,
                        parameterValues.get(methodParameter).get(parameterOffset.get(methodParameter))));
            }
            incrementOffset(parameterOffset, parameterValues);
            result.add(resultRow);
        }
        return result;
    }

    private void incrementOffset(Map<MethodParameter, Integer> parameterOffset,
                                 Map<MethodParameter, List<Number>> parameterValues) {
        List<MethodParameter> parameters = new ArrayList<>(parameterOffset.keySet());
        int i = 0;
        while (i < parameters.size() && !isAllValuesUsed(parameterOffset, parameterValues)) {
            if (parameterOffset.get(parameters.get(i)) == parameterValues.get(parameters.get(i)).size() - 1) {
                parameterOffset.put(parameters.get(i), 0);
                i++;
                continue;
            }
            if (parameterOffset.get(parameters.get(i)) < parameterValues.get(parameters.get(i)).size()) {
                parameterOffset.put(parameters.get(i), parameterOffset.get(parameters.get(i)) + 1);
                return;
            }
            i++;
        }
    }

    private boolean isAllValuesUsed(Map<MethodParameter, Integer> parameterOffset,
                                    Map<MethodParameter, List<Number>> parameterValues) {
        for (MethodParameter methodParameter : parameterOffset.keySet()) {
            if (parameterOffset.get(methodParameter) != parameterValues.get(methodParameter).size() - 1) {
                return false;
            }
        }
        return true;
    }
}
